((nil
  (indent-tabs-mode . t)
  (tab-width . 8))
 (c-mode
  (c-file-style . "bsd")
  (c-basic-offset . 8))
 (shell-mode
  (sh-basic-offset . 8)
  (sh-indentation . 8))
 (python-mode
  (indent-tabs-mode . nil))
 )
